use std::{fs, path::{PathBuf, Path}, process::Command};

use anyhow::Context as _;
use clap::Parser;

use crate::build_ebpf::{self, build_ebpf, Architecture};

#[derive(Debug, Parser)]
pub struct Options {
    /// Set the endianness of the BPF target
    #[clap(default_value = "bpfel-unknown-none", long)]
    pub bpf_target: Architecture,
    // Output folder
    #[clap(short, long, default_value = "bin")]
    pub output_folder: PathBuf,
}

fn build_user(_opts: &Options) -> Result<(), anyhow::Error> {
    let status = Command::new("cargo")
        .args(["build","--package=responder", "--release", "--features=default_artifact_build"])
        .status()
        .expect("failed to build responder userspace program");
    if !status.success() {
        return Err(anyhow::anyhow!(
            "cargo build --release failed to run: {}",
            status.to_string()
        ));
    }
    Ok(())
}

fn build_tools(_opts: &Options) -> Result<(), anyhow::Error> {
    let status = Command::new("cargo")
        .args(["build", "--package=responder-tools", "--release"])
        .status()
        .expect("failed to build responder tools");
    if !status.success() {
        return Err(anyhow::anyhow!(
            "cargo build --release failed to run: {}",
            status.to_string()
        ));
    }
    Ok(())
}

fn copy_execs_from_dir<P: AsRef<Path>, Q: AsRef<Path>>(from: P, to: Q) -> Result<(), anyhow::Error>{
    for entry in fs::read_dir(from)? {
        if let Ok(entry) = entry {
            let path = entry.path();
            if path.is_file()
                && !path
                    .file_name()
                    .unwrap()
                    .to_str()
                    .map(|s| s.contains("."))
                    .unwrap_or(true)
            {
                let to = to.as_ref().join(path.file_name().unwrap());
                fs::copy(&path, &to).context(format!(
                    "Error copying from {} to {}",
                    path.to_str().unwrap_or("[path not valid unicode]"),
                    to.to_str().unwrap_or("[path not valid unicode]")
                ))?;
            }
        }
    }
    Ok(())
}


pub fn build(opts: Options) -> Result<(), anyhow::Error> {
    fs::create_dir_all(&opts.output_folder).context(format!(
        "Error creating dirs for path: {}",
        opts.output_folder
            .to_str()
            .unwrap_or("[path not valid unicode]")
    ))?;

    build_ebpf(build_ebpf::Options {
        target: opts.bpf_target,
        release: true,
    })
    .context("Error while building eBPF program")?;

    let ebpf_bin = opts.output_folder.join(r"ebpf");
    fs::create_dir_all(&ebpf_bin)?;
    copy_execs_from_dir(
        [r"target", opts.bpf_target.to_string().as_str(), r"release"]
            .iter()
            .collect::<PathBuf>(),
        ebpf_bin
    )?;

    build_user(&opts).context("Error while building userspace application")?;
    fs::copy(
        "target/release/responder",
        opts.output_folder.join("responder"),
    )?;

    build_tools(&opts).context("Error while building tool applications")?;

    let tools_bin = opts.output_folder.join(r"tools");
    fs::create_dir_all(&tools_bin)?;
    copy_execs_from_dir("target/release", tools_bin)?;

    Ok(())
}
